package org.processmining.mixedparadigm.algorithms.replayer;

import gnu.trove.iterator.TShortIterator;
import gnu.trove.list.TIntList;
import gnu.trove.list.TShortList;
import gnu.trove.list.array.TIntArrayList;
import gnu.trove.set.TShortSet;
import gnu.trove.set.hash.TShortHashSet;

import java.util.ArrayList;
import java.util.HashMap;
import java.util.List;
import java.util.Map;

import nl.tue.astar.AStarThread;
import nl.tue.astar.Trace;
import nl.tue.astar.impl.State;
import nl.tue.astar.impl.memefficient.CachedStorageAwareDelegate;
import nl.tue.astar.impl.memefficient.TailInflater;
import nl.tue.astar.util.ShortShortMultiset;
import nl.tue.storage.CompressedHashSet;
import nl.tue.storage.CompressedStore;
import nl.tue.storage.Deflater;
import nl.tue.storage.StorageException;

import org.deckfour.xes.classification.XEventClass;
import org.deckfour.xes.classification.XEventClasses;
import org.deckfour.xes.model.XLog;
import org.processmining.mixedparadigm.models.dfa.DeterministicFiniteAutomaton;
import org.processmining.mixedparadigm.models.mixedparadigm.DeclarativeArc;
import org.processmining.mixedparadigm.models.mixedparadigm.PetrinetWithAutomata;
import org.processmining.mixedparadigm.models.mixedparadigm.constraints.ConstraintType;
import org.processmining.models.graphbased.directed.petrinet.elements.Transition;
import org.processmining.models.semantics.petrinet.Marking;
import org.processmining.plugins.astar.petrinet.impl.AbstractPDelegate;
import org.processmining.plugins.astar.petrinet.impl.AbstractPILPDelegate;
import org.processmining.plugins.astar.petrinet.impl.PHead;
import org.processmining.plugins.astar.petrinet.impl.PHeadCompressor;
import org.processmining.plugins.connectionfactories.logpetrinet.TransEvClassMapping;

public class PILPDelegateAutomata extends AbstractPILPDelegate<PAutomataTail> implements
		CachedStorageAwareDelegate<PHead, PAutomataTail> {

	protected PAutomataTailCompressor tailCompressor;

	private List<DeterministicFiniteAutomaton> automata;

	private int[] constraint2Cost;
	private DeclarativeArc[] short2constraint;

	private CompressedHashSet<State<PHead, PAutomataTail>> statespace;

	private boolean allowViolatingConstraints;

	public PILPDelegateAutomata(PetrinetWithAutomata net, XLog log, XEventClasses classes, TransEvClassMapping map,
			Map<Transition, Integer> mapTrans2Cost, Map<XEventClass, Integer> mapEvClass2Cost,
			Map<DeclarativeArc, Integer> mapConstraintToCost, boolean allowViolatingConstraints, int delta,
			int threads, Marking... set) {
		this(net, log, classes, map, mapTrans2Cost, mapEvClass2Cost, new HashMap<Transition, Integer>(0),
				mapConstraintToCost, allowViolatingConstraints, delta, threads, set);
	}

	/**
	 * The following constructors accept mapping from sync moves to cost
	 */
	public PILPDelegateAutomata(PetrinetWithAutomata net, XLog log, XEventClasses classes, TransEvClassMapping map,
			Map<Transition, Integer> mapTrans2Cost, Map<XEventClass, Integer> mapEvClass2Cost,
			Map<Transition, Integer> mapSync2Cost, Map<DeclarativeArc, Integer> mapConstraintToCost,
			boolean allowViolatingConstraints, int delta, int threads, Marking... set) {
		super(net, log, classes, map, mapTrans2Cost, mapEvClass2Cost, mapSync2Cost, delta, threads, true,
				allowViolatingConstraints, set);
		this.allowViolatingConstraints = allowViolatingConstraints;

		this.automata = new ArrayList<>();
		this.constraint2Cost = new int[net.getConstraints().size()];
		short2constraint = new DeclarativeArc[net.getConstraints().size()];

		int a = 0;
		for (DeclarativeArc constraint : net.getConstraints()) {
			ConstraintType type = constraint.getType();
			if (type instanceof ConstraintType.Unary) {
				automata.add(((ConstraintType.Unary) type).getAutomaton(trans2int, constraint.getSource()));
			} else if (type instanceof ConstraintType.Binary) {
				automata.add(((ConstraintType.Binary) type).getAutomaton(trans2int, constraint.getSource(),
						constraint.getTarget()));
			} else if (type instanceof ConstraintType.Branched) {
				automata.add(((ConstraintType.Branched) type).getAutomaton(trans2int, constraint.getSources(),
						constraint.getTargets()));
			}
			constraint2Cost[a] = mapConstraintToCost.get(constraint);
			short2constraint[a] = constraint;
			a++;
		}

		System.out.println("Allow violation: " + allowViolatingConstraints);

		// make sure automata are complete by adding a PV state
		for (a = 0; a < automata.size(); a++) {
			if (allowViolatingConstraints) {
				automata.get(a).complete();
			} else {
				automata.get(a).reduce();
			}
		}

		((PAutomataHeadCompressor<PAutomataTail>) this.headCompressor).setAutomataCount((short) automata.size());
		this.tailCompressor = new PAutomataTailCompressor(2 * transitions + activities + resetArcs + set.length,
				places, activities, (short) automata.size());

	}

	@Override
	public void setStateSpace(CompressedHashSet<State<PHead, PAutomataTail>> statespace) {
		this.statespace = statespace;
	}

	//BVD: Issue still open is the case termination if we allow for violation of constraints

	public TIntList getEnabledTransitionsChangingMarking(ShortShortMultiset marking, short[] automataStates) {
		TIntList list = new TIntArrayList();
		if (automataStates[automataStates.length - 1] != 0) {
			// we terminated, nothing else to do...
			return list;
		}

		for (short t = 0; t < transitions; t++) {
			if (isEnabled(t, marking, automataStates)) {

				ShortShortMultiset newMarking = constructNewMarking(marking, t);

				boolean changes = !newMarking.equals(marking);

				if (!getTransition(t).isInvisible()) {
					// check if in at least one automaton the state changes
					for (int a = 0; !changes && a < automata.size(); a++) {
						short nextState = automata.get(a).getNextState(automataStates[a], t);
						if (nextState != automataStates[a]) {
							changes = true;
						}
					}
				}
				if (changes) {
					if (finalMarkings.contains(newMarking)) {
						// we can reach a final marking from this state.
						// add the terminating move as well
						list.add(-t - 1);
					}
					// add the regular move.
					list.add(t);
				}
			}
		}
		return list;
	}

	public TIntList getSyncMoves(ShortShortMultiset marking, short[] states, int activity) {
		// only consider transitions mapped to activity
		final TIntList result = new TIntArrayList();
		if (states[states.length - 1] != 0) {
			// we terminated, nothing else to do...
			return result;
		}

		TShortList trans = getTransitions((short) activity);
		TShortIterator it = trans.iterator();
		while (it.hasNext()) {
			short i = it.next();
			if (isEnabled(i, marking, states)) {
				ShortShortMultiset newMarking = constructNewMarking(marking, i);
				result.add(i);
				if (finalMarkings.contains(newMarking)) {
					// include the case-terminating final state
					result.add(-i - 1);
				}
			}
		}

		return result;
	}

	public boolean isEnabled(int transition, final ShortShortMultiset marking, short[] automataStates) {
		boolean en = super.isEnabled(transition, marking);
		// termination flag not triggered.
		en &= automataStates[automataStates.length - 1] == 0;
		if (!allowViolatingConstraints && !getTransition((short) transition).isInvisible()) {
			// for visible transitions, check the automata
			for (int a = 0; en && a < automata.size(); a++) {
				en &= automata.get(a).isAllowed(automataStates[a], transition);
			}
		}

		return en;
	}

	public TShortSet getConstraintsBeingViolated(boolean terminating, short transition, short[] automataStates) {
		// if violating constraints is not allowed, then this method returns false.
		TShortSet constraints = new TShortHashSet();

		if (!getTransition(transition).isInvisible()) {
			// for visible transitions, check the automata
			for (short a = 0; a < automata.size(); a++) {

				if (terminating && temporarilyViolatesConstraint(a, transition, automataStates[a])
				// if terminating and the constraint is TV after execution of modelMove, OR
						|| permanentlyViolatesConstraint(a, transition, automataStates[a])) {
					// if the constraint is PV after execution of modelMove (and not before)
					constraints.add(a);
				}

			}
		}

		return constraints;
	}

	public boolean permanentlyViolatesConstraint(short constraint, int transition, short currentState) {
		if (getTransition((short) transition).isInvisible()) {
			return false;
		}

		// fromstate is NOT already permanently violated 
		return automata.get(constraint).getStateType(currentState) != DeterministicFiniteAutomaton.PV &&
		// AND toState is permanently violated
				automata.get(constraint).getStateType(automata.get(constraint).getNextState(currentState, transition)) == DeterministicFiniteAutomaton.PV;

	}

	public boolean temporarilyViolatesConstraint(short constraint, int transition, short currentState) {
		if (getTransition((short) transition).isInvisible()) {
			return false;
		}
		if (!automata.get(constraint).isAllowed(currentState, transition)) {
			return true;
		}
		// fromstate is NOT already temporarily violated 
		return automata.get(constraint).getStateType(automata.get(constraint).getNextState(currentState, transition)) == DeterministicFiniteAutomaton.TV;

	}

	public boolean isFinal(ShortShortMultiset marking, short[] automataStates) {
		boolean done = finalMarkings.contains(marking);

		if (allowViolatingConstraints) {
			done &= automataStates[automataStates.length - 1] == 1;
		} else {
			for (int a = 0; done && a < automata.size(); a++) {
				if (!automata.get(a).isAcceptingState(automataStates[a])) {
					// non-accepting state, allow only if allowing to violate constraints
					done = false;
				}
			}
		}

		return done;
	}

	public short getNextState(int automaton, short currentState, int modelMove) {
		if (getTransition((short) modelMove).isInvisible()) {
			// for invisible transitions, dont change state
			return currentState;
		} else {
			return automata.get(automaton).getNextState(currentState, modelMove);
		}
	}

	@Override
	protected PHeadCompressor<PAutomataTail> constructHeadCompressor(short places, short activities) {
		return new PAutomataHeadCompressor<PAutomataTail>(places, activities);
	}

	@Override
	public MPRecord createInitialRecord(PHead head, Trace trace) {
		MPRecord r = new MPRecord(0, null, head.getMarking().getNumElts(), trace.getSize());
		r.setEstimatedRemainingCost(0, true);
		return r;
	}

	public short[] getInitialStatesAutomata() {
		short[] initial = new short[automata.size() + 1];
		for (int i = 0; i < automata.size(); i++) {
			initial[i] = automata.get(i).getInitialState();
		}

		return initial;
	}

	public boolean isAllowingViolationOfConstraints() {
		return allowViolatingConstraints;
	}

	public PAutomataTail createInitialTail(PHead head) {
		return new PAutomataTail(this, head, 0);
	}

	public PHeadCompressor<PAutomataTail> getHeadInflater() {
		return headCompressor;
	}

	public PHeadCompressor<PAutomataTail> getHeadDeflater() {
		return headCompressor;
	}

	public TailInflater<PAutomataTail> getTailInflater() {
		return tailCompressor;
	}

	public Deflater<PAutomataTail> getTailDeflater() {
		return tailCompressor;
	}

	public int getCostFor(long currentState, PAutomataHead nextHead, int modelMove, int activity) {

		if (!allowViolatingConstraints) {
			return super.getCostFor(modelMove, activity);
		}
		// add costs for violating automata
		PAutomataHead head;
		try {
			head = ((PAutomataHead) getStorage().getObject(currentState).getHead());
		} catch (StorageException e) {
			// don't know what to do here.
			throw new RuntimeException(e);
		}

		short[] automataStates = head.getStates();

		boolean terminating = nextHead.getStates()[nextHead.getStates().length - 1] != 0;
		if (terminating) {
			assert modelMove == AStarThread.NOMOVE || modelMove < 0;
			if (modelMove != AStarThread.NOMOVE) {
				modelMove = -modelMove - 1;
			}

		} else {
			assert modelMove == AStarThread.NOMOVE || modelMove >= 0;
		}

		// If terminating, this is a transition into a final state

		if (modelMove == AStarThread.NOMOVE) {
			return getCostForMoveLog(terminating, automataStates, (short) activity);
		}

		// there was a move in the model and violation of constraints is allowed.
		// If this transition goes into the final marking, then
		// we need to consider the temporarily violated constraints 
		// as permanently violated.

		if (activity == AStarThread.NOMOVE) {
			return getCostForMoveModel(terminating, automataStates, (short) modelMove);
		} else {
			// synchronous move assumed here
			return getCostForMoveSync(terminating, automataStates, (short) modelMove);
		}
	}

	public int getCostForMoveLog(boolean terminating, short[] automataStates, short activity) {
		int cost = super.getCostForMoveLog(activity);
		assert allowViolatingConstraints;
		if (terminating) {
			// Final logMove ends up in terminating state. Add costs for
			// TV constraints.
			for (short a = 0; a < automata.size(); a++) {
				if (isTemporarilyViolated(a, automataStates[a])) {
					cost += getDelta() * constraint2Cost[a];
				}
			}
		}

		return cost;
	}

	private boolean isTemporarilyViolated(short a, short s) {
		return automata.get(a).getStateType(s) == DeterministicFiniteAutomaton.TV;
	}

	public int getCostForMoveSync(boolean terminating, short[] automataStates, short modelMove) {
		return super.getCostForMoveSync(modelMove) + getCostForConstraints(terminating, automataStates, modelMove);
	}

	public int getCostForMoveModel(boolean terminating, short[] automataStates, short modelMove) {
		return super.getCostForMoveModel(modelMove) + getCostForConstraints(terminating, automataStates, modelMove);
	}

	public int getCostForConstraints(boolean terminating, short[] automataStates, short modelMove) {
		int cost = 0;
		for (short a = 0; a < automata.size(); a++) {
			if (terminating && temporarilyViolatesConstraint(a, modelMove, automataStates[a])) {
				// if terminating and the constraint is TV after execution of modelMove, account costs
				cost += getDelta() * constraint2Cost[a];
			} else if (permanentlyViolatesConstraint(a, modelMove, automataStates[a])) {
				// if the constraint is PV after execution of modelMove (and note before), account costs
				cost += getDelta() * constraint2Cost[a];
			}
		}
		return cost;
	}

	public CompressedStore<State<PHead, PAutomataTail>> getStorage() {
		return statespace.getBackingStore();
	}

	public DeclarativeArc getDeclarativeArc(short a) {
		return short2constraint[a];
	}

	public ShortShortMultiset constructNewMarking(ShortShortMultiset marking, short transition) {
		ShortShortMultiset newMarking = marking.clone();
		if (transition < 0) {
			transition = (short) (-transition - 1);
		}
		// clone the marking
		short[] input = transIndex2input.get(transition);
		short[] output = transIndex2output.get(transition);

		for (short place = numPlaces(); place-- > 0;) {
			short val = newMarking.get(place);
			short needed = input[place];
			assert (needed <= val);
			if (needed != AbstractPDelegate.INHIBITED) {
				// only adjust the value for non-inhibitor arcs
				newMarking.adjustValue(place, (short) -needed);
			}
		}

		for (short place = numPlaces(); place-- > 0;) {
			short val = newMarking.get(place);
			short produced = output[place];
			if (produced < 0) {
				// combination or reset arc and regular arc (regular arc may be 0)
				// first get the actual produced tokens
				produced = (short) (-(produced + 1));
				// then account for removing all tokens first
				produced -= val;
			}
			newMarking.adjustValue(place, produced);
		}
		return newMarking;
	}

	/**
	 * update the states of the automata
	 * 
	 * @param states
	 * @param modelMove
	 */
	public void updateAutomataStates(short[] states, short modelMove) {
		for (int a = 0; a < automata.size(); a++) {
			if (modelMove < 0) {
				// this is a case-terminating transition.
				// put the state into a negative number...
				states[a] = getNextState(a, states[a], (short) (-modelMove - 1));
				states[states.length - 1] = 1;
			} else {
				states[a] = getNextState(a, states[a], modelMove);
			}
		}
	}

	@Override
	public TShortSet getActivitiesFor(short transition) {
		if (transition < 0) {
			return super.getActivitiesFor((short) (-transition - 1));
		} else {
			return super.getActivitiesFor(transition);
		}
	}

	@Override
	public Transition getTransition(short i) {
		if (i < 0) {
			return super.getTransition((short) (-i - 1));
		} else {
			return super.getTransition(i);
		}
	}

	public int getNumAutomata() {
		return automata.size();
	}

}
