package org.processmining.mixedparadigm.algorithms.replayer;

import java.util.ArrayList;
import java.util.HashMap;
import java.util.HashSet;
import java.util.List;
import java.util.Map;
import java.util.Set;

import org.deckfour.xes.classification.XEventClass;
import org.deckfour.xes.classification.XEventClasses;
import org.deckfour.xes.info.XLogInfo;
import org.deckfour.xes.info.XLogInfoFactory;
import org.deckfour.xes.model.XEvent;
import org.deckfour.xes.model.XLog;
import org.deckfour.xes.model.XTrace;
import org.processmining.framework.plugin.PluginContext;
import org.processmining.framework.plugin.annotations.KeepInProMCache;
import org.processmining.mixedparadigm.models.mixedparadigm.DeclarativeArc;
import org.processmining.mixedparadigm.models.mixedparadigm.PetrinetWithAutomata;
import org.processmining.mixedparadigm.parameters.MPReplayParameterProvider;
import org.processmining.mixedparadigm.parameters.MPReplayParameters;
import org.processmining.models.graphbased.directed.petrinet.PetrinetGraph;
import org.processmining.models.graphbased.directed.petrinet.elements.Transition;
import org.processmining.models.semantics.petrinet.Marking;
import org.processmining.plugins.astar.petrinet.AbstractPetrinetReplayer;
import org.processmining.plugins.astar.petrinet.impl.AbstractPDelegate;
import org.processmining.plugins.astar.petrinet.impl.PRecord;
import org.processmining.plugins.connectionfactories.logpetrinet.TransEvClassMapping;
import org.processmining.plugins.petrinet.replayer.algorithms.IPNReplayParameter;
import org.processmining.plugins.petrinet.replayer.algorithms.costbasedcomplete.CostBasedCompleteParam;
import org.processmining.plugins.petrinet.replayer.annotations.PNReplayAlgorithm;
import org.processmining.plugins.petrinet.replayresult.PNRepResult;
import org.processmining.plugins.petrinet.replayresult.StepTypes;

import gnu.trove.iterator.TIntIterator;
import gnu.trove.iterator.TShortIterator;
import gnu.trove.list.TIntList;
import gnu.trove.map.TIntIntMap;
import gnu.trove.set.TShortSet;
import nl.tue.astar.AStarThread;
import nl.tue.astar.Trace;

@KeepInProMCache
@PNReplayAlgorithm
public class PetrinetReplayerWithAutomata extends AbstractPetrinetReplayer<PAutomataTail, PILPDelegateAutomata> {

	private Map<DeclarativeArc, Integer> mapConstr2Cost;
	private boolean allowViolations;

	/**
	 * Return true if input of replay without parameters are correct
	 */
	public boolean isReqWOParameterSatisfied(PluginContext context, PetrinetGraph net, XLog log,
			TransEvClassMapping mapping) {
		if (net instanceof PetrinetWithAutomata) {
			// check number of transitions, places, and event classes, should be less than Short.MAX_VALUE
			if ((net.getTransitions().size() < Short.MAX_VALUE) && (net.getPlaces().size() < Short.MAX_VALUE)) {
				// check the number of event classes, should be less than Short.MAX_VALUE
				XLogInfo summary = XLogInfoFactory.createLogInfo(log, mapping.getEventClassifier());
				return (summary.getEventClasses().getClasses().size() < Short.MAX_VALUE);
			}
		}
		return false;
	}

	/**
	 * Return true if all replay inputs are correct
	 */
	public boolean isAllReqSatisfied(PluginContext context, PetrinetGraph net, XLog log, TransEvClassMapping mapping,
			IPNReplayParameter parameter) {
		if (net instanceof PetrinetWithAutomata) {
			// check number of transitions, places, and event classes, should be less than Short.MAX_VALUE
			if ((net.getTransitions().size() < Short.MAX_VALUE) && (net.getPlaces().size() < Short.MAX_VALUE)) {
				// check the number of event classes, should be less than Short.MAX_VALUE
				XLogInfo summary = XLogInfoFactory.createLogInfo(log, mapping.getEventClassifier());
				if (summary.getEventClasses().getClasses().size() < Short.MAX_VALUE) {
					return isParameterReqCorrect(net, log, mapping, parameter);
				}

			}
		}
		return false;
	}

	public String toString() {
		return "A* Cost-based Fitness Express with ILP and automata, assuming at most " + Short.MAX_VALUE
				+ " tokens in each place.";
	}

	protected PILPDelegateAutomata getDelegate(PetrinetGraph net, XLog log, XEventClasses classes,
			TransEvClassMapping mapping, int delta, int threads) {
		for(Transition t: mapping.keySet()){
			System.out.println(t+" "+mapping.get(t));
		}
		if (net instanceof PetrinetWithAutomata) {
			return new PILPDelegateAutomata((PetrinetWithAutomata) net, log, classes, mapping, mapTrans2Cost,
					mapEvClass2Cost, mapSync2Cost, mapConstr2Cost, allowViolations, delta, threads, finalMarkings);
		}
		return null;
	}

	@Override
	protected PAutomataHead constructHead(PILPDelegateAutomata delegate, Marking m, XTrace xtrace) {
		return new PAutomataHead(delegate, m, xtrace, delegate.getInitialStatesAutomata());
	}

//	protected SyncReplayResult recordToResult(AbstractPDelegate<?> d, XTrace trace, Trace filteredTrace, PRecord r,
//			int traceIndex, int stateCount, boolean isReliable, long milliseconds, int queuedStates, int traversedArcs,
//			int minCostMoveModel, TIntList unUsedIndices, TIntIntMap trace2orgTrace) {
	@Override
	protected MPSyncReplayResult recordToResult(AbstractPDelegate<?> d, XTrace trace, Trace filteredTrace, PRecord r,
			int traceIndex, int stateCount, boolean isReliable, long milliseconds, int queuedStates, int traversedArcs,
			int minCostMoveModel, TIntList unUsedIndices, TIntIntMap trace2orgTrace) {
		PILPDelegateAutomata da = (PILPDelegateAutomata) d;

		List<MPRecord> history = MPRecord.getHistory((MPRecord) r);
		double mmCost = 0; // total cost of move on model
		double mlCost = 0; // total cost of move on log
		double mSyncCost = 0; // total cost of synchronous move

		double mmUpper = 0; // total cost if all movements are move on model (including the synchronous one)
		double mlUpper = 0; // total cost if all events are move on log

		int eventInTrace = -1;
		List<StepTypes> stepTypes = new ArrayList<StepTypes>(history.size());
		List<Object> nodeInstance = new ArrayList<Object>();
		List<Set<DeclarativeArc>> violatedConstraints = new ArrayList<Set<DeclarativeArc>>();
		Set<DeclarativeArc> emptyList = new HashSet<>(0);

		TIntIterator it = unUsedIndices.iterator();
		int firstUnUsed = it.hasNext() ? it.next() : Integer.MAX_VALUE;

		short[] automataStates = da.getInitialStatesAutomata();

		// History is the ordered sequence of PRecords that correspond to the alignment
		for (int ri = 0; ri < history.size(); ri++) {
			MPRecord rec = history.get(ri);
			short mm = (short) rec.getModelMove();
			if (mm < 0) {
				mm = (short) (-mm - 1);
			}
			if (rec.getMovedEvent() == AStarThread.NOMOVE) {

				// move model only
				Transition t = d.getTransition(mm);

				if (t.isInvisible()) {
					stepTypes.add(StepTypes.MINVI);
				} else {
					stepTypes.add(StepTypes.MREAL);
				}
				nodeInstance.add(t);
				mmCost += (da.getCostForMoveModel(rec.getModelMove() < 0, automataStates, mm) - 1.0) / d.getDelta();
				mmUpper += (da.getCostForMoveModel(rec.getModelMove() < 0, automataStates, mm) - 1.0) / d.getDelta();
			} else {
				// a move occurred in the log. Check if class aligns with class in trace

				// check rec.getMovedEvent. If this is larger than unUsedIndices, then include all unUsedIndices
				// upto rec.getMovedEvent as LogMoves right now.
				while (trace2orgTrace.get(rec.getMovedEvent()) > firstUnUsed) {
					XEventClass clsInTrace = d.getClassOf(trace.get(firstUnUsed)); // this an unused event

					stepTypes.add(StepTypes.L);
					nodeInstance.add(clsInTrace);
					mlCost += mapEvClass2Cost.get(clsInTrace);
					eventInTrace++;

					firstUnUsed = it.hasNext() ? it.next() : Integer.MAX_VALUE;
				}

				short a = (short) filteredTrace.get(rec.getMovedEvent()); // a is the event obtained from the replay
				eventInTrace++;

				if (rec.getModelMove() == AStarThread.NOMOVE) {
					// move log only
					stepTypes.add(StepTypes.L);
					nodeInstance.add(d.getEventClass(a));
					mlCost += (d.getCostForMoveLog(a) - 1.0) / d.getDelta();
					//					mlUpper += (d.getCostForMoveLog(a) - 1.0) / d.getDelta();
				} else {
					// sync move
					// check if constraint violated
					stepTypes.add(StepTypes.LMGOOD);
					nodeInstance.add(d.getTransition(mm));
					mSyncCost += (da.getCostForMoveSync(rec.getModelMove() < 0, automataStates, mm) - 1.0)
							/ d.getDelta();
					mmUpper += (da.getCostForMoveModel(rec.getModelMove() < 0, automataStates, mm) - 1.0)
							/ d.getDelta();
				}
			}
			Set<DeclarativeArc> violated = emptyList;
			if (rec.getModelMove() != AStarThread.NOMOVE) {
				TShortSet violatedAutomata = da.getConstraintsBeingViolated(isReliable && ri == history.size() - 1, mm,
						automataStates);
				violated = new HashSet<>(violatedAutomata.size());
				TShortIterator vaIt = violatedAutomata.iterator();
				while (vaIt.hasNext()) {
					violated.add(da.getDeclarativeArc(vaIt.next()));
				}
				for (short a = 0; a < da.getNumAutomata(); a++) {
					automataStates[a] = da.getNextState(a, automataStates[a], mm);
				}
			}
			if (violated.isEmpty()) {
				violatedConstraints.add(emptyList);
			} else {
				violatedConstraints.add(violated);
			}

		}

		// add the rest of the trace
		eventInTrace++;
		//		while (eventInTrace < trace.size()) {
		while (firstUnUsed < trace.size()) {
			// move log only
			XEventClass a = d.getClassOf(trace.get(firstUnUsed));
			eventInTrace++;
			stepTypes.add(StepTypes.L);
			nodeInstance.add(a);
			violatedConstraints.add(emptyList);

			System.out.println("A: "+a.toString()+" and "+trace.get(firstUnUsed).getAttributes().get("concept:name"));
			mlCost += mapEvClass2Cost.get(a);
			//			mlUpper += mapEvClass2Cost.get(a);
			firstUnUsed = it.hasNext() ? it.next() : Integer.MAX_VALUE;

		}

		// calculate mlUpper (because in cases where we have synchronous move in manifest, more than one events are aggregated
		// in one movement
		for (XEvent evt : trace) {
			mlUpper += mapEvClass2Cost.get(d.getClassOf(evt));
		}

		MPSyncReplayResult res = new MPSyncReplayResult(nodeInstance, stepTypes, traceIndex, violatedConstraints);

		res.setReliable(isReliable);
		Map<String, Double> info = new HashMap<String, Double>();
		info.put(PNRepResult.RAWFITNESSCOST, (mmCost + mlCost + mSyncCost));

		if (mlCost > 0) {
			info.put(PNRepResult.MOVELOGFITNESS, 1 - (mlCost / mlUpper));
		} else {
			info.put(PNRepResult.MOVELOGFITNESS, 1.0);
		}

		if (mmCost > 0) {
			info.put(PNRepResult.MOVEMODELFITNESS, 1 - (mmCost / mmUpper));
		} else {
			info.put(PNRepResult.MOVEMODELFITNESS, 1.0);
		}
		info.put(PNRepResult.NUMSTATEGENERATED, (double) stateCount);
		info.put(PNRepResult.QUEUEDSTATE, (double) queuedStates);

		// set info fitness
		if (mmCost > 0 || mlCost > 0 || mSyncCost > 0) {
			info.put(PNRepResult.TRACEFITNESS, 1 - ((mmCost + mlCost + mSyncCost) / (mlUpper + minCostMoveModel)));
		} else {
			info.put(PNRepResult.TRACEFITNESS, 1.0);
		}
		info.put(PNRepResult.TIME, (double) milliseconds);
		info.put(PNRepResult.ORIGTRACELENGTH, (double) eventInTrace);
		res.setInfo(info);
		return res;
	}

	public String getHTMLInfo() {
		return "<html>This is an algorithm to calculate cost-based fitness between a log and a Petri net with automata. <br/><br/>"
				+ "Given a trace and a Petri net, this algorithm "
				+ "return a matching between the trace and an allowed firing sequence of the net with the"
				+ "least deviation cost. The firing sequence has to reach proper "
				+ "termination of the net, specified by 1 final marking.</html>";
	}

	public MPReplayParameterProvider constructParamProvider(PluginContext context, PetrinetGraph net, XLog log,
			TransEvClassMapping mapping) {
		return new MPReplayParameterProvider(context, (PetrinetWithAutomata) net, log, mapping);
	}

	@Override
	protected void importParameters(CostBasedCompleteParam parameters) {
		super.importParameters(parameters);
		mapConstr2Cost = ((MPReplayParameters) parameters).getMapConstraint2Cost();
		allowViolations = ((MPReplayParameters) parameters).isAllowViolatingConstraints();
	}

}