package org.processmining.mixedparadigm.models.dfa;

public interface DeterministicFiniteAutomaton {

	// state types Permanently or Temporarily Violated or Satisfied
	public final static byte PV = 0x00; //000
	public final static byte TV = 0x04; //100
	public final static byte PS = 0x03; //011
	public final static byte TS = 0x07; //111
	public final static byte AC = 0x01; //001

	/**
	 * Adds a new state to the matrix
	 * 
	 * @param isInitial
	 * @param type
	 * @return
	 */
	public abstract short addState(boolean isInitial, byte type, boolean addSelfLoops);

	/**
	 * Add a transition from state fromS to state toS labeled with label
	 * 
	 * @param fromS
	 * @param toS
	 * @param label
	 */
	public abstract void addTransition(short fromS, short toS, short label);

	/**
	 * Remove transition labeled with label from state fromS
	 * 
	 * @param fromS
	 * @param toS
	 * @param label
	 */
	public abstract void removeTransition(short fromS, short label);

	/**
	 * returns the new state after executing a transition with the given label
	 * in currentState, or NOSTATE is no such new state exists.
	 * 
	 * @param currentState
	 * @param label
	 * @return
	 */
	public abstract short getNextState(short currentState, int label);

	/**
	 * Checks if the transition with the given label is possible in the given
	 * state
	 * 
	 * @param currentState
	 * @param label
	 * @return
	 */
	public abstract boolean isAllowed(short currentState, int label);

	/**
	 * Returns the initial state
	 * 
	 * @return
	 */
	public abstract short getInitialState();

	/**
	 * Sets the initial state
	 * 
	 * @param state
	 */
	public abstract void setInitialState(short state);

	/**
	 * Returns true if the given state is an accepting state
	 * 
	 * @param state
	 * @return
	 */
	public abstract boolean isAcceptingState(short state);

	public abstract void setStateType(short state, byte type);

	/**
	 * Reduces the DFA to remove all permanently violated states by simply
	 * removing all transitions into these states and then removing unreachable
	 * states.
	 */
	public abstract void reduce();

	/**
	 * Adds a single PV type state to which all other states can transition
	 * using any label not currently in the DFA. After this call, all states
	 * have an equal amount of outgoing transitions, equal to the alphabet.
	 */
	public abstract void complete();

	/**
	 * Check if the automaton is complete, i.e. for all states, there is a
	 * transition defined for all labels
	 * 
	 * @return
	 */
	public abstract boolean isComplete();

	/**
	 * Returns the type of the state (PV, TV, PS, or TS)
	 * 
	 * @param state
	 * @return
	 */
	public abstract byte getStateType(short state);

	public abstract boolean equals(Object o);

	public abstract int hashCode();

	public abstract String toString();

	public abstract String getLabel();

}