package org.processmining.mixedparadigm.models.dfa;

import java.util.Arrays;

/**
 * A memory inefficient implementation of DFA's, but effective for small model
 * sizes and low number of transitions.
 * 
 * The automaton should be deterministic and complete, i.e. any activity that is
 * not in the set of outgoing transitions of a state is simply not allowed to
 * occur.
 * 
 * @author bfvdonge
 * 
 */
public class GeneralDeterministicFiniteAutomaton implements DeterministicFiniteAutomaton {

	private final static short NOSTATE = Short.MIN_VALUE;

	protected final int alphabetSize;

	/**
	 * Stores the representation of this automaton. This is a very dense
	 * representation. Each row represents a state, each column a
	 * transitionlabel. A transition from state s to state t labelled with a is
	 * encoded as matrix[s][a] = t
	 */
	protected short[][] matrix;

	private short states;

	protected short initialState;

	protected final String label;

	/**
	 * Constructs a DFA over a given alphabet from [0..alphabetSize) (exclusive)
	 * without states.
	 * 
	 * @param alphabetSize
	 */
	public GeneralDeterministicFiniteAutomaton(String label, int alphabetSize) {
		this(label, alphabetSize, 10);
	}

	/**
	 * Constructs a DFA over a given alphabet from [0..alphabetSize) (exclusive)
	 * without states.
	 * 
	 * @param alphabetSize
	 */
	public GeneralDeterministicFiniteAutomaton(String label, int alphabetSize, int matrixSize) {
		this.label = label;
		assert alphabetSize < Short.MAX_VALUE;
		this.alphabetSize = alphabetSize;
		matrix = new short[matrixSize + 1][];
		states = 0;
	}

	/*
	 * (non-Javadoc)
	 * 
	 * @see
	 * org.processmining.mixedparadigm.models.DeterministicFiniteAutomaton#addState
	 * (boolean, byte, boolean)
	 */
	@Override
	public short addState(boolean isInitial, byte type, boolean addSelfLoops) {
		// check if array needs to grow
		if (states == matrix.length) {
			short[][] newMatrix = new short[states + 10][];
			for (int i = 0; i < states; i++) {
				newMatrix[i] = matrix[i];
			}
			matrix = newMatrix;
		}
		// create state
		matrix[states] = new short[alphabetSize + 1];
		// initialize the transitions
		if (addSelfLoops) {
			Arrays.fill(matrix[states], states);
		} else {
			Arrays.fill(matrix[states], NOSTATE);
		}
		// set the state type
		setStateType(states, type);
		// set initial state
		if (isInitial) {
			setInitialState(states);
		}

		return states++;
	}

	/*
	 * (non-Javadoc)
	 * 
	 * @see org.processmining.mixedparadigm.models.DeterministicFiniteAutomaton#
	 * addTransition(short, short, short)
	 */
	@Override
	public void addTransition(short fromS, short toS, short label) {
		matrix[fromS][label] = toS;
	}

	/*
	 * (non-Javadoc)
	 * 
	 * @see org.processmining.mixedparadigm.models.DeterministicFiniteAutomaton#
	 * removeTransition(short, short)
	 */
	@Override
	public void removeTransition(short fromS, short label) {
		addTransition(fromS, NOSTATE, label);
	}

	/*
	 * (non-Javadoc)
	 * 
	 * @see org.processmining.mixedparadigm.models.DeterministicFiniteAutomaton#
	 * getNextState(short, int)
	 */
	@Override
	public short getNextState(short currentState, int label) {
		return matrix[currentState][label];
	}

	/*
	 * (non-Javadoc)
	 * 
	 * @see
	 * org.processmining.mixedparadigm.models.DeterministicFiniteAutomaton#isAllowed
	 * (short, int)
	 */
	@Override
	public boolean isAllowed(short currentState, int label) {
		return getNextState(currentState, label) != NOSTATE;
	}

	/*
	 * (non-Javadoc)
	 * 
	 * @see org.processmining.mixedparadigm.models.DeterministicFiniteAutomaton#
	 * getInitialState()
	 */
	@Override
	public short getInitialState() {
		return initialState;
	}

	/*
	 * (non-Javadoc)
	 * 
	 * @see org.processmining.mixedparadigm.models.DeterministicFiniteAutomaton#
	 * setInitialState(short)
	 */
	@Override
	public void setInitialState(short state) {
		initialState = state;
	}

	/*
	 * (non-Javadoc)
	 * 
	 * @see org.processmining.mixedparadigm.models.DeterministicFiniteAutomaton#
	 * isAcceptingState(short)
	 */
	@Override
	public boolean isAcceptingState(short state) {
		// check accepting state bit
		return (matrix[state][alphabetSize] & AC) == AC;
	}

	/*
	 * (non-Javadoc)
	 * 
	 * @see org.processmining.mixedparadigm.models.DeterministicFiniteAutomaton#
	 * setStateType(short, byte)
	 */
	@Override
	public void setStateType(short state, byte type) {
		matrix[state][alphabetSize] = type;
	}

	/*
	 * (non-Javadoc)
	 * 
	 * @see
	 * org.processmining.mixedparadigm.models.DeterministicFiniteAutomaton#reduce
	 * ()
	 */
	@Override
	public void reduce() {
		for (short s = states; s-- > 0;) {
			if (getStateType(s) == PV) {
				// shift states up
				for (short t = 0; t < states - 1; t++) {
					for (short i = 0; i < alphabetSize; i++) {
						if (matrix[t][i] == s) {
							// transition into s. remove
							matrix[t][i] = NOSTATE;
						} else if (matrix[t][i] > s) {
							// transition from t to a s'>s. Removing s, 
							// so reduce by 1
							matrix[t][i]--;
						}
					}
					if (t > s) {
						matrix[t - 1] = matrix[t];
					}
				}
				states--;
			}
		}
	}

	/*
	 * (non-Javadoc)
	 * 
	 * @see
	 * org.processmining.mixedparadigm.models.DeterministicFiniteAutomaton#complete
	 * ()
	 */
	@Override
	public void complete() {
		short newState = -1;
		for (short s = states; s-- > 0;) {
			for (short l = 0; l < alphabetSize; l++) {
				if (matrix[s][l] == NOSTATE) {
					// no transition labeled l from state s
					if (newState == -1) {
						newState = addState(false, PV, true);
					}
					matrix[s][l] = newState;
				}
			}
		}
		if (newState != -1) {
			for (short l = 0; l < alphabetSize; l++) {
				matrix[states - 1][l] = newState;
			}
		}
	}

	/*
	 * (non-Javadoc)
	 * 
	 * @see org.processmining.mixedparadigm.models.DeterministicFiniteAutomaton#
	 * isComplete()
	 */
	@Override
	public boolean isComplete() {
		for (short s = states; s-- > 0;) {
			for (short l = 0; l < alphabetSize; l++) {
				if (matrix[s][l] == NOSTATE) {
					return false;
				}
			}
		}
		return true;
	}

	/*
	 * (non-Javadoc)
	 * 
	 * @see org.processmining.mixedparadigm.models.DeterministicFiniteAutomaton#
	 * getStateType(short)
	 */
	@Override
	public byte getStateType(short state) {
		return (byte) (matrix[state][alphabetSize] & 0x00FF);
	}

	/*
	 * (non-Javadoc)
	 * 
	 * @see
	 * org.processmining.mixedparadigm.models.DeterministicFiniteAutomaton#equals
	 * (java.lang.Object)
	 */
	@Override
	public boolean equals(Object o) {
		return (o instanceof GeneralDeterministicFiniteAutomaton)
				&& Arrays.deepEquals(matrix, ((GeneralDeterministicFiniteAutomaton) o).matrix);
	}

	/*
	 * (non-Javadoc)
	 * 
	 * @see
	 * org.processmining.mixedparadigm.models.DeterministicFiniteAutomaton#hashCode
	 * ()
	 */
	@Override
	public int hashCode() {
		return Arrays.deepHashCode(matrix);
	}

	/*
	 * (non-Javadoc)
	 * 
	 * @see
	 * org.processmining.mixedparadigm.models.DeterministicFiniteAutomaton#toString
	 * ()
	 */
	@Override
	public String toString() {
		return Arrays.deepToString(matrix);
	}

	public String getLabel() {
		return label;
	}
}
