package org.processmining.mixedparadigm.models.mixedparadigm.constraints;

import gnu.trove.map.TObjectShortMap;

import java.util.Collection;

import org.processmining.mixedparadigm.models.dfa.DeterministicFiniteAutomaton;
import org.processmining.mixedparadigm.models.mixedparadigm.Graphics;
import org.processmining.mixedparadigm.models.mixedparadigm.constraints.binary.AlternatePrecedence;
import org.processmining.mixedparadigm.models.mixedparadigm.constraints.binary.AlternateResponse;
import org.processmining.mixedparadigm.models.mixedparadigm.constraints.binary.AlternateSuccession;
import org.processmining.mixedparadigm.models.mixedparadigm.constraints.binary.ChainPrecedence;
import org.processmining.mixedparadigm.models.mixedparadigm.constraints.binary.ChainResponse;
import org.processmining.mixedparadigm.models.mixedparadigm.constraints.binary.ChainSuccession;
import org.processmining.mixedparadigm.models.mixedparadigm.constraints.binary.CoExistence;
import org.processmining.mixedparadigm.models.mixedparadigm.constraints.binary.NotChainSuccession;
import org.processmining.mixedparadigm.models.mixedparadigm.constraints.binary.NotCoExistence;
import org.processmining.mixedparadigm.models.mixedparadigm.constraints.binary.NotSuccession;
import org.processmining.mixedparadigm.models.mixedparadigm.constraints.binary.Precedence;
import org.processmining.mixedparadigm.models.mixedparadigm.constraints.binary.RespondedExistence;
import org.processmining.mixedparadigm.models.mixedparadigm.constraints.binary.Response;
import org.processmining.mixedparadigm.models.mixedparadigm.constraints.binary.Succession;
import org.processmining.mixedparadigm.models.mixedparadigm.constraints.branched.BranchedAlternatePrecedence;
import org.processmining.mixedparadigm.models.mixedparadigm.constraints.branched.BranchedPrecedence;
import org.processmining.mixedparadigm.models.mixedparadigm.constraints.branched.BranchedRespondedExistence;
import org.processmining.mixedparadigm.models.mixedparadigm.constraints.branched.BranchedResponse;
import org.processmining.mixedparadigm.models.mixedparadigm.constraints.branched.Choice;
import org.processmining.mixedparadigm.models.mixedparadigm.constraints.branched.ExclusiveChoice;
import org.processmining.mixedparadigm.models.mixedparadigm.constraints.unary.AbsenceN;
import org.processmining.mixedparadigm.models.mixedparadigm.constraints.unary.ExactlyN;
import org.processmining.mixedparadigm.models.mixedparadigm.constraints.unary.ExistenceN;
import org.processmining.mixedparadigm.models.mixedparadigm.constraints.unary.Init;
import org.processmining.models.graphbased.AttributeMap.ArrowType;
import org.processmining.models.graphbased.directed.petrinet.elements.Transition;

public interface ConstraintType {

	static final String INIT = "Init";
	static final String ATLEAST_ = "At Least ";
	static final String ATMOST_ = "At Most ";
	static final String EXACT_ = "Exactly ";

	static final String RESPONSE = "Response";
	static final String NOTCHAINSUC = "Not Chain Succession";
	static final String ALTERNATERESPONSE = "Alternate Response";
	static final String PRECEDENCE = "Precedence";
	static final String RESPONDEDEXISTENCE = "Responded Existence";
	static final String SUCCESSION = "Succession";
	static final String COEXISTENCE = "Coexistence";
	static final String CHAINSUCCESSION = "Chain Succession";
	static final String CHAINRESPONSE = "Chain Response";
	static final String CHAINPRECEDENCE = "Chain Precedence";
	static final String ALTERNATESUC = "Alternate Succession";
	static final String ALTERNATEPRECEDENCE = "Alternate Precedence";
	static final String NOTSUC = "Not Succession";
	static final String NOTCOEXIST = "Not Coexistence";

	// branched constraints
	static final String EXCLUSIVECHOICE_ = "Exclusive Choice ";
	static final String CHOICE_ = "Choice ";
	static final String BRANCHEDRESPONSE = "Branched Response";
	static final String BRANCHEDPRECEDENCE = "Branched Precedence";
	static final String BRANCHEDRESPONDEDEXISTENCE = "Branched Responded Existence";
	static final String BRANCHEDALTERNATEPRECEDENCE = "Branched Alternate Precedence";

	/**
	 * Helper class to construct constraints based on their name.
	 * 
	 * @author bfvdonge
	 * 
	 */
	public final static class Helper {

		public static ConstraintType fromName(String name) {
			// Binary constraints without parameters
			if (name.equals(RESPONSE))
				return new Response();
			if (name.equals(INIT))
				return new Init();
			if (name.equals(NOTCHAINSUC))
				return new NotChainSuccession();
			if (name.equals(ALTERNATERESPONSE))
				return new AlternateResponse();
			if (name.equals(PRECEDENCE))
				return new Precedence();
			if (name.equals(RESPONDEDEXISTENCE))
				return new RespondedExistence();
			if (name.equals(SUCCESSION))
				return new Succession();
			if (name.equals(COEXISTENCE))
				return new CoExistence();
			if (name.equals(CHAINSUCCESSION))
				return new ChainSuccession();
			if (name.equals(CHAINRESPONSE))
				return new ChainResponse();
			if (name.equals(CHAINPRECEDENCE))
				return new ChainPrecedence();
			if (name.equals(ALTERNATESUC))
				return new AlternateSuccession();
			if (name.equals(ALTERNATEPRECEDENCE))
				return new AlternatePrecedence();
			if (name.equals(NOTSUC))
				return new NotSuccession();
			if (name.equals(NOTCOEXIST))
				return new NotCoExistence();

			// Branched constraints
			if (name.equals(BRANCHEDPRECEDENCE))
				return new BranchedPrecedence();
			if (name.equals(BRANCHEDRESPONSE))
				return new BranchedResponse();
			if (name.equals(BRANCHEDRESPONDEDEXISTENCE))
				return new BranchedRespondedExistence();
			if (name.equals(BRANCHEDALTERNATEPRECEDENCE))
				return new BranchedAlternatePrecedence();

			short val;
			try {
				val = Short.parseShort(name.substring(name.lastIndexOf(" ") + 1));
			} catch (Exception e) {
				return null;
			}

			// unary constraints with parameters
			if (name.startsWith(ATLEAST_))
				return new ExistenceN(val);
			if (name.startsWith(ATMOST_))
				return new AbsenceN(++val);
			if (name.startsWith(EXACT_))
				return new ExactlyN(val);

			// N-ary constraints with parameter
			//			if (name.startsWith(EXCLUSIVECHOICE_))
			//				return new ExclusiveChoice(val);
			if (name.startsWith(CHOICE_))
				return new Choice(val);
			if (name.startsWith(EXCLUSIVECHOICE_))
				return new ExclusiveChoice(val);

			// UNKNOWN CONSTRAINT TYPE
			assert false;
			return null;
		}
	}

	/**
	 * Returns the name of this constraint
	 * 
	 * @return
	 */
	public String getName();

	/**
	 * Returns the label used in the visualization of this constraint
	 * 
	 * @return
	 */
	public String getLabel();

	public Graphics getGraphics();

	abstract static class Unary implements ConstraintType {
		protected Graphics graphics = new Graphics(1, ArrowType.ARROWTYPE_NONE, ArrowType.ARROWTYPE_NONE,
				ArrowType.ARROWTYPE_NONE, true);

		public Graphics getGraphics() {
			return graphics;
		}

		/**
		 * Returns the automaton representation of this constraint type.
		 * 
		 * @param trans2int
		 * @param from
		 * @param to
		 * @return
		 */
		public abstract DeterministicFiniteAutomaton getAutomaton(TObjectShortMap<Transition> trans2int, Transition from);

		public String getLabel() {
			return getName();
		}

		public ActivationMode getActivationMode() {
			return ActivationMode.GLOBAL;
		}

	}

	static abstract class Binary implements ConstraintType {
		protected Graphics graphics = new Graphics(1, ArrowType.ARROWTYPE_NONE, ArrowType.ARROWTYPE_NONE,
				ArrowType.ARROWTYPE_NONE, false);
		private final ActivationMode mode;

		public Binary(ActivationMode mode) {
			this.mode = mode;

		}

		public Graphics getGraphics() {
			return graphics;
		}

		/**
		 * Returns the automaton representation of this constraint type.
		 * 
		 * @param trans2int
		 * @param from
		 * @param to
		 * @return
		 */
		public abstract DeterministicFiniteAutomaton getAutomaton(TObjectShortMap<Transition> trans2int,
				Transition from, Transition to);

		public String getLabel() {
			return getName();
		}

		public ActivationMode getActivationMode() {
			return mode;
		}

	}

	static abstract class Branched implements ConstraintType {
		protected Graphics graphics = new Graphics(1, ArrowType.ARROWTYPE_NONE, ArrowType.ARROWTYPE_NONE,
				ArrowType.ARROWTYPE_NONE, false);

		private final ActivationMode mode;

		public Branched(ActivationMode mode) {
			this.mode = mode;

		}

		public Graphics getGraphics() {
			return graphics;
		}

		/**
		 * Returns the automaton representation of this constraint type.
		 * 
		 * @param trans2int
		 * @param from
		 * @param to
		 * @return
		 */
		public abstract DeterministicFiniteAutomaton getAutomaton(TObjectShortMap<Transition> trans2int,
				Collection<? extends Transition> from, Collection<? extends Transition> to);

		public String getLabel() {
			return getName();
		}

		public ActivationMode getActivationMode() {
			return mode;
		}
	}

	public ActivationMode getActivationMode();

}
