package org.processmining.mixedparadigm.plugins.pnml;

import java.awt.geom.Point2D;
import java.util.ArrayList;
import java.util.Collection;
import java.util.Iterator;
import java.util.List;
import java.util.Map;

import org.processmining.framework.util.Pair;
import org.processmining.mixedparadigm.models.mixedparadigm.DeclarativeArc;
import org.processmining.models.connections.GraphLayoutConnection;
import org.processmining.models.graphbased.AbstractGraphElement;
import org.processmining.models.graphbased.directed.DirectedGraphElement;
import org.processmining.models.graphbased.directed.petrinet.elements.ExpandableSubNet;
import org.processmining.models.graphbased.directed.petrinet.elements.Transition;
import org.processmining.models.jgraph.DirectedGraphEdgeForMultiGraph;
import org.processmining.models.jgraph.DirectedGraphNodeForMultiGraph;
import org.processmining.plugins.pnml.base.Pnml;
import org.xmlpull.v1.XmlPullParser;

public class DPnmlArcGraphics extends DPnmlGraphics {

	/**
	 * PNML annotation graphics tag.
	 */
	public final static String TAG = "multiEdgeGraphics";

	/**
	 * graphics of multi-edges.
	 */
	private final List<DPnmlMultiArcGraphics> arcGraphicsList;
	/**
	 * graphics of intermediate node.
	 */
	private DPnmlMultiNodeGraphics nodeGraphics;

	/**
	 * Creates a fresh PNML arc graphics.
	 */
	protected DPnmlArcGraphics() {
		super(TAG);
		arcGraphicsList = new ArrayList<DPnmlMultiArcGraphics>();
		nodeGraphics = null;
	}

	/**
	 * Checks whether the current start tag is known. If known, it imports the
	 * corresponding child element and returns true. Otherwise, it returns
	 * false.
	 * 
	 * @return Whether the start tag was known.
	 */
	protected boolean importElements(XmlPullParser xpp, Pnml pnml) {
		if (super.importElements(xpp, pnml)) {
			return true;
		}
		if (xpp.getName().equals(DPnmlMultiNodeGraphics.TAG)) {
			/*
			 * Line element.
			 */
			nodeGraphics = ((DPnmlElementFactory) factory).createDPnmlMultiNodeGraphics();
			nodeGraphics.importElement(xpp, pnml);
			return true;
		}
		if (xpp.getName().equals(DPnmlMultiArcGraphics.TAG)) {
			/*
			 * Position element.
			 */
			DPnmlMultiArcGraphics arcGraphics = ((DPnmlElementFactory) factory).createDPnmlMultiArcGraphics();
			arcGraphics.importElement(xpp, pnml);
			arcGraphicsList.add(arcGraphics);
			return true;
		}
		/*
		 * Unknown start tag.
		 */
		return false;
	}

	/**
	 * Exports the arc graphics.
	 */
	protected String exportElements(Pnml pnml) {
		String s = super.exportElements(pnml);
		if (nodeGraphics != null) {
			s += nodeGraphics.exportElement(pnml);
		}
		for (DPnmlMultiArcGraphics arcGraphics : arcGraphicsList) {
			s += arcGraphics.exportElement(pnml);
		}
		return s;
	}

	/**
	 * Sets the graphics for the given graph element.
	 * 
	 * @param subNet
	 *            The given sub net.
	 * @param element
	 *            The given element.
	 * @param displacement
	 *            The displacement for this sub net.
	 */
	public void convertToNet(ExpandableSubNet subNet, DeclarativeArc element, Map<String, Transition> transitionMap,
			Point2D.Double displacement, GraphLayoutConnection layout) {
		List<DirectedGraphElement> elements = new ArrayList<>(element.getSources().size() + element.getTargets().size()
				+ 1);

		DirectedGraphNodeForMultiGraph node;
		if (nodeGraphics != null) {
			node = new DirectedGraphNodeForMultiGraph(element);
			elements.add(node);
			nodeGraphics.convertToNet(subNet, node, displacement, layout);

			// now the edges
			for (DPnmlMultiArcGraphics arcGraphics : arcGraphicsList) {
				DirectedGraphEdgeForMultiGraph edge;
				if (arcGraphics.toIntermediate) {
					edge = new DirectedGraphEdgeForMultiGraph(transitionMap.get(arcGraphics.otherNode), node, element);
				} else {
					edge = new DirectedGraphEdgeForMultiGraph(node, transitionMap.get(arcGraphics.otherNode), element);
				}
				arcGraphics.convertToNet(subNet, edge, displacement, layout);
				elements.add(edge);
			}
		}
		layout.setMultiGraphElements(element, elements);
	}

	public DPnmlArcGraphics convertFromNet(ExpandableSubNet parent, DeclarativeArc element,
			Map<Pair<AbstractGraphElement, ExpandableSubNet>, String> idMap, GraphLayoutConnection layout) {
		DPnmlArcGraphics result = null;
		try {
			Collection<DirectedGraphElement> elements = layout.getMultiGraphElements(element);
			if (elements == null) {
				return this;
			}
			Iterator<DirectedGraphElement> it = elements.iterator();
			nodeGraphics = ((DPnmlElementFactory) factory).createDPnmlMultiNodeGraphics().convertFromNet(parent,
					(DirectedGraphNodeForMultiGraph) it.next(), layout);
			while (it.hasNext()) {
				arcGraphicsList.add(((DPnmlElementFactory) factory).createDPnmlMultiArcGraphics().convertFromNet(
						parent, (DirectedGraphEdgeForMultiGraph) it.next(), idMap, layout));
			}

			if (!arcGraphicsList.isEmpty() || (nodeGraphics != null)) {
				result = this;
			}
		} catch (Exception ex) {
		}
		return result;
	}
}
