package org.processmining.mixedparadigm.plugins.pnml;

import java.awt.geom.Point2D;
import java.util.ArrayList;
import java.util.List;
import java.util.Map;

import org.processmining.framework.util.Pair;
import org.processmining.models.connections.GraphLayoutConnection;
import org.processmining.models.graphbased.AbstractGraphElement;
import org.processmining.models.graphbased.directed.DirectedGraphNode;
import org.processmining.models.graphbased.directed.petrinet.elements.ExpandableSubNet;
import org.processmining.models.jgraph.DirectedGraphEdgeForMultiGraph;
import org.processmining.plugins.pnml.base.Pnml;
import org.processmining.plugins.pnml.base.PnmlElement;
import org.processmining.plugins.pnml.elements.graphics.PnmlPosition;
import org.xmlpull.v1.XmlPullParser;

public class DPnmlMultiArcGraphics extends PnmlElement {

	/**
	 * PNML annotation graphics tag.
	 */
	public final static String TAG = "edgegraphics";

	/**
	 * Positions elements (may be multiple).
	 */
	protected final List<PnmlPosition> positionList;
	/**
	 * to Intermediate element.
	 */
	protected boolean toIntermediate;
	/**
	 * other node element.
	 */
	protected String otherNode;

	/**
	 * Creates a fresh PNML arc graphics.
	 */
	protected DPnmlMultiArcGraphics() {
		this(TAG);
	}

	/**
	 * Creates a fresh PNML arc graphics.
	 */
	protected DPnmlMultiArcGraphics(String tag) {
		super(tag);
		positionList = new ArrayList<PnmlPosition>();
		otherNode = null;
		toIntermediate = false;
	}

	/**
	 * Imports all known attributes.
	 */
	protected void importAttributes(XmlPullParser xpp, Pnml pnml) {
		/*
		 * Import known basic object attributes.
		 */
		super.importAttributes(xpp, pnml);
		/*
		 * Import otherNode attribute.
		 */
		importOtherNode(xpp, pnml);
		/*
		 * Import target attribute.
		 */
		importToIntermediate(xpp, pnml);
	}

	private void importOtherNode(XmlPullParser xpp, Pnml pnml) {
		String value = xpp.getAttributeValue(null, "otherNode");
		if (value != null) {
			otherNode = value;
		}
	}

	private String exportOtherNode(Pnml pnml) {
		if (otherNode != null) {
			return exportAttribute("otherNode", otherNode, pnml);
		}
		return "";
	}

	private void importToIntermediate(XmlPullParser xpp, Pnml pnml) {
		String value = xpp.getAttributeValue(null, "toIntermediate");
		if (value != null) {
			toIntermediate = Boolean.parseBoolean(value);
		}
	}

	private String exportToIntermediate(Pnml pnml) {
		return exportAttribute("toIntermediate", Boolean.toString(toIntermediate), pnml);
	}

	/**
	 * Exports all attributes.
	 */
	protected String exportAttributes(Pnml pnml) {
		return super.exportAttributes(pnml) + exportOtherNode(pnml) + exportToIntermediate(pnml);
	}

	/**
	 * Checks whether the current start tag is known. If known, it imports the
	 * corresponding child element and returns true. Otherwise, it returns
	 * false.
	 * 
	 * @return Whether the start tag was known.
	 */
	protected boolean importElements(XmlPullParser xpp, Pnml pnml) {
		if (super.importElements(xpp, pnml)) {
			return true;
		}
		if (xpp.getName().equals(PnmlPosition.TAG)) {
			/*
			 * Position element.
			 */
			PnmlPosition position = factory.createPnmlPosition();
			position.importElement(xpp, pnml);
			positionList.add(position);
			return true;
		}
		/*
		 * Unknown start tag.
		 */
		return false;
	}

	/**
	 * Exports the arc graphics.
	 */
	protected String exportElements(Pnml pnml) {
		String s = super.exportElements(pnml);
		for (PnmlPosition position : positionList) {
			s += position.exportElement(pnml);
		}
		return s;
	}

	/**
	 * Sets the graphics for the given graph element.
	 * 
	 * @param subNet
	 *            The given sub net.
	 * @param element
	 *            The given element.
	 * @param displacement
	 *            The displacement for this sub net.
	 */
	public void convertToNet(ExpandableSubNet subNet, AbstractGraphElement element, Point2D.Double displacement,
			GraphLayoutConnection layout) {
		ArrayList<Point2D> points = new ArrayList<Point2D>();
		layout.setEdgePoints(element, points);
		for (PnmlPosition position : positionList) {
			position.convertToNet(subNet, element, points, displacement);
		}
	}

	public DPnmlMultiArcGraphics convertFromNet(ExpandableSubNet parent, DirectedGraphEdgeForMultiGraph element,
			Map<Pair<AbstractGraphElement, ExpandableSubNet>, String> idMap, GraphLayoutConnection layout) {
		DPnmlMultiArcGraphics result = null;
		try {
			List<Point2D> points = layout.getEdgePoints(element);
			for (Point2D point : points) {
				positionList.add(factory.createPnmlPosition().convertFromNet(parent, point));
			}
			toIntermediate = element.isToIntermediate();

			Pair<DirectedGraphNode, ExpandableSubNet> pair = new Pair<>(toIntermediate ? element.getSource()
					: element.getTarget(), parent);
			otherNode = idMap.get(pair);
			if (otherNode != null) {
				result = this;
			}
		} catch (Exception ex) {
		}
		return result;
	}

}
