package org.processmining.mixedparadigm.plugins.pnml;

import java.awt.geom.Point2D;
import java.util.ArrayList;
import java.util.Iterator;
import java.util.List;
import java.util.Map;

import org.jgraph.graph.GraphConstants;
import org.processmining.framework.util.Pair;
import org.processmining.mixedparadigm.models.mixedparadigm.DeclarativeArc;
import org.processmining.mixedparadigm.models.mixedparadigm.PetrinetWithAutomata;
import org.processmining.mixedparadigm.models.mixedparadigm.constraints.ConstraintType;
import org.processmining.models.connections.GraphLayoutConnection;
import org.processmining.models.graphbased.AbstractGraphElement;
import org.processmining.models.graphbased.AttributeMap;
import org.processmining.models.graphbased.directed.DirectedGraphElement;
import org.processmining.models.graphbased.directed.petrinet.PetrinetEdge;
import org.processmining.models.graphbased.directed.petrinet.PetrinetGraph;
import org.processmining.models.graphbased.directed.petrinet.PetrinetNode;
import org.processmining.models.graphbased.directed.petrinet.elements.ExpandableSubNet;
import org.processmining.models.graphbased.directed.petrinet.elements.Place;
import org.processmining.models.graphbased.directed.petrinet.elements.Transition;
import org.processmining.plugins.pnml.base.Pnml;
import org.processmining.plugins.pnml.elements.PnmlBasicObject;
import org.processmining.plugins.pnml.elements.PnmlPage;
import org.processmining.plugins.pnml.toolspecific.PnmlToolSpecific;
import org.xmlpull.v1.XmlPullParser;

public class PnmlDeclarativeArc extends PnmlBasicObject {

	/**
	 * PNML arc tag.
	 */
	public final static String TAG = "declarativeArc";

	/**
	 * Id attribute.
	 */
	private String id;
	/**
	 * Source attribute.
	 */
	private List<DPnmlSourceTransitionReference> sources;
	/**
	 * Target attribute.
	 */
	private List<DPnmlTargetTransitionReference> targets;

	/**
	 * Constraint name.
	 */
	private ConstraintType constraint;

	/**
	 * Graphics element.
	 */
	private DPnmlGraphics graphics;

	/**
	 * Creates a fresh PNML arc.
	 */
	protected PnmlDeclarativeArc() {
		super(TAG);
		id = null;
		sources = new ArrayList<>(1);
		targets = new ArrayList<>(1);
		graphics = null;
		constraint = null;
	}

	/**
	 * Imports all known attributes.
	 */
	protected void importAttributes(XmlPullParser xpp, Pnml pnml) {
		/*
		 * Import known basic object attributes.
		 */
		super.importAttributes(xpp, pnml);
		/*
		 * Import id attribute.
		 */
		importId(xpp, pnml);
		/*
		 * Import target attribute.
		 */
		importName(xpp, pnml);
	}

	private void importName(XmlPullParser xpp, Pnml pnml) {
		String value = xpp.getAttributeValue(null, "constraintName");
		if (value != null) {
			constraint = ConstraintType.Helper.fromName(value);
		}
	}

	/**
	 * Exports all attributes.
	 */
	protected String exportAttributes(Pnml pnml) {
		return super.exportAttributes(pnml) + exportId(pnml) + exportName(pnml);
	}

	private void importId(XmlPullParser xpp, Pnml pnml) {
		String value = xpp.getAttributeValue(null, "id");
		if (value != null) {
			id = value;
		}
	}

	private String exportId(Pnml pnml) {
		if (id != null) {
			return exportAttribute("id", id, pnml);
		}
		return "";
	}

	private String exportName(Pnml pnml) {
		if (constraint != null) {
			return exportAttribute("constraintName", constraint.getName(), pnml);
		}
		return "";
	}

	/**
	 * Checks whether the current start tag is known. If known, it imports the
	 * corresponding child element and returns true. Otherwise, it returns
	 * false.
	 * 
	 * @return Whether the start tag was known.
	 */
	protected boolean importElements(XmlPullParser xpp, Pnml pnml) {
		if (super.importElements(xpp, pnml)) {
			/*
			 * Start tag corresponds to a known child element of a PNML basic
			 * object.
			 */
			return true;
		}
		if (xpp.getName().equals(DPnmlSingleArcGraphics.TAG)) {
			/*
			 * Graphics element. create graphics object and import graphics
			 * element.
			 */
			if (graphics != null) {
				return false;
			}
			graphics = ((DPnmlElementFactory) factory).createDPnmlSingleArcGraphics();
			graphics.importElement(xpp, pnml);
			return true;
		}
		if (xpp.getName().equals(DPnmlArcGraphics.TAG)) {
			/*
			 * Graphics element. create graphics object and import graphics
			 * element.
			 */
			if (graphics != null) {
				return false;
			}
			graphics = ((DPnmlElementFactory) factory).createDPnmlArcGraphics();
			graphics.importElement(xpp, pnml);
			return true;
		}
		if (xpp.getName().equals(DPnmlSourceTransitionReference.TAG)) {
			// source transition
			DPnmlSourceTransitionReference ref = ((DPnmlElementFactory) factory).createDPnmlSourceTransitionReference();
			ref.importElement(xpp, pnml);
			sources.add(ref);
			return true;
		}
		if (xpp.getName().equals(DPnmlTargetTransitionReference.TAG)) {
			// source transition
			DPnmlTargetTransitionReference ref = ((DPnmlElementFactory) factory).createDPnmlTargetTransitionReference();
			ref.importElement(xpp, pnml);
			targets.add(ref);
			return true;
		}

		return false;
	}

	/**
	 * Exports all child elements.
	 */
	protected String exportElements(Pnml pnml) {
		String s = super.exportElements(pnml);
		for (DPnmlSourceTransitionReference ref : sources) {
			s += ref.exportElement(pnml);
		}
		for (DPnmlTargetTransitionReference ref : targets) {
			s += ref.exportElement(pnml);
		}
		if (graphics != null) {
			s += graphics.exportElement(pnml);
		}
		return s;
	}

	/**
	 * Checks validity. Should have a source and a target.
	 */
	protected void checkValidity(Pnml pnml) {
		super.checkValidity(pnml);
		if ((id == null) || (sources.isEmpty()) || (targets.isEmpty()) || (constraint == null)) {
			pnml.log(tag, lineNumber,
					"Expected id, at least one source, at least one target and constraintName of known constraint");
		}
	}

	/**
	 * Converts this PNML arc to a declarative arc
	 * 
	 * @param net
	 *            The net to add the arc to.
	 * @param subNet
	 *            The sub net to add the arc to.
	 * @param placeMap
	 *            The places found so far.
	 * @param transitionMap
	 *            The transitions found so far.
	 */
	public void convertToNet(PetrinetGraph net, ExpandableSubNet subNet, Map<String, Place> placeMap,
			Map<String, Transition> transitionMap,
			Map<String, PetrinetEdge<? extends PetrinetNode, ? extends PetrinetNode>> edgeMap,
			Point2D.Double displacement, GraphLayoutConnection layout) {
		/*
		 * Create arc (if source and target can be found).
		 */
		DeclarativeArc arc = null;

		Iterator<DPnmlSourceTransitionReference> sIt = sources.iterator();
		Iterator<DPnmlTargetTransitionReference> tIt = targets.iterator();

		String source = sIt.next().getName("");
		String target = tIt.next().getName("");

		assert (transitionMap.containsKey(source) && transitionMap.containsKey(target));

		ExpandableSubNet sourceSubNet = transitionMap.get(source).getParent();
		ExpandableSubNet targetSubNet = transitionMap.get(target).getParent();

		arc = ((PetrinetWithAutomata) net).addConstraint(transitionMap.get(source), transitionMap.get(target),
				constraint, firstCommonAncestor(sourceSubNet, targetSubNet));

		while (sIt.hasNext()) {
			arc.addAdditionalSource(transitionMap.get(sIt.next().getName("")));
		}
		while (tIt.hasNext()) {
			arc.addAdditionalTarget(transitionMap.get(tIt.next().getName("")));
		}
		/*
		 * If arc created, set graphics and inscription.
		 */
		if (arc != null) {
			edgeMap.put(id, arc);
			arc.getAttributeMap().put(AttributeMap.STYLE, GraphConstants.STYLE_SPLINE);
			super.convertToNet(arc);
			for (PnmlToolSpecific toolSpecific : toolSpecificList) {
				toolSpecific.convertToNet(arc);
			}
			if (graphics != null) {
				if (graphics instanceof DPnmlArcGraphics) {
					((DPnmlArcGraphics) graphics).convertToNet(subNet, arc, transitionMap, displacement, layout);
				} else {
					((DPnmlSingleArcGraphics) graphics).convertToNet(subNet, arc, displacement, layout);
				}
			}

		}
	}

	private ExpandableSubNet firstCommonAncestor(ExpandableSubNet subNet1, ExpandableSubNet subNet2) {
		int depth1 = 0;
		int depth2 = 0;
		for (ExpandableSubNet subNet = subNet1; subNet != null; subNet = subNet.getParent()) {
			depth1++;
		}
		for (ExpandableSubNet subNet = subNet2; subNet != null; subNet = subNet.getParent()) {
			depth2++;
		}
		while (depth1 > depth2) {
			subNet1 = subNet1.getParent();
			depth1--;
		}
		while (depth1 < depth2) {
			subNet2 = subNet2.getParent();
			depth2--;
		}
		while ((subNet1 != null) && (subNet1 != subNet2)) {
			subNet1 = subNet1.getParent();
			subNet2 = subNet2.getParent();
		}
		return subNet1;
	}

	public PnmlDeclarativeArc convertFromNet(ExpandableSubNet parent, DeclarativeArc edge, PnmlPage page,
			Map<Pair<AbstractGraphElement, ExpandableSubNet>, String> idMap, GraphLayoutConnection layout) {
		super.convertFromNet(edge.getLabel());
		PnmlToolSpecific toolSpecific = factory.createPnmlToolSpecific();
		toolSpecific.convertFromNet(edge);
		toolSpecificList.add(toolSpecific);

		id = "constraint" + idMap.size();
		idMap.put(new Pair<AbstractGraphElement, ExpandableSubNet>(edge, parent), id);

		for (Transition sourceNode : edge.getSources()) {
			if (!idMap.containsKey(new Pair<DirectedGraphElement, ExpandableSubNet>(sourceNode, parent))) {
				/*
				 * Source is in different subnet than the edge. Need to create a
				 * reference to the source in the edge's subnet (page) and use
				 * that reference as source.
				 */
				page.convertFromNet(parent, sourceNode, idMap, layout);
			}
			String id = idMap.get(new Pair<DirectedGraphElement, ExpandableSubNet>(sourceNode, parent));
			DPnmlSourceTransitionReference ref = ((DPnmlElementFactory) factory).createDPnmlSourceTransitionReference();
			ref.convertFromNet(id);
			sources.add(ref);
		}

		for (Transition targetNode : edge.getTargets()) {
			if (!idMap.containsKey(new Pair<DirectedGraphElement, ExpandableSubNet>(targetNode, parent))) {
				/*
				 * Target is in different subnet than the edge. Need to create a
				 * reference to the target in the edge's subnet (page) and use
				 * that reference as target.
				 */
				page.convertFromNet(parent, targetNode, idMap, layout);
			}
			String id = idMap.get(new Pair<DirectedGraphElement, ExpandableSubNet>(targetNode, parent));
			DPnmlTargetTransitionReference ref = ((DPnmlElementFactory) factory).createDPnmlTargetTransitionReference();
			ref.convertFromNet(id);
			targets.add(ref);
		}
		constraint = edge.getType();

		if (edge.getSources().size() > 1 || edge.getTargets().size() > 1) {
			// declarative multi-arcs require special attention.
			graphics = ((DPnmlElementFactory) factory).createDPnmlArcGraphics().convertFromNet(parent, edge, idMap,
					layout);
		} else {
			graphics = ((DPnmlElementFactory) factory).createDPnmlSingleArcGraphics().convertFromNet(parent, edge,
					layout);
		}
		return this;
	}
}