package org.processmining.mixedparadigm.visualisation;

import gnu.trove.map.TObjectIntMap;
import gnu.trove.map.hash.TObjectIntHashMap;

import java.util.ArrayList;
import java.util.Collections;
import java.util.Comparator;
import java.util.List;
import java.util.Set;

import org.deckfour.xes.model.XLog;
import org.processmining.mixedparadigm.algorithms.replayer.MPReplayResult;
import org.processmining.mixedparadigm.algorithms.replayer.MPSyncReplayResult;
import org.processmining.mixedparadigm.models.mixedparadigm.DeclarativeArc;
import org.processmining.mixedparadigm.models.mixedparadigm.PetrinetWithAutomata;
import org.processmining.mixedparadigm.models.mixedparadigm.constraints.ActivationMode;
import org.processmining.models.graphbased.directed.petrinet.PetrinetEdge;
import org.processmining.models.graphbased.directed.petrinet.PetrinetNode;
import org.processmining.models.graphbased.directed.petrinet.elements.Transition;
import org.processmining.models.semantics.petrinet.Marking;
import org.processmining.plugins.connectionfactories.logpetrinet.TransEvClassMapping;
import org.processmining.plugins.pnalignanalysis.visualization.projection.CoreInfoProvider;
import org.processmining.plugins.replayer.replayresult.SyncReplayResult;

public class MPCoreInfoProvider extends CoreInfoProvider {

	protected int[] activations;
	protected int[] activatedTraces;
	protected int[] activationsFitting;
	protected int[] violations;
	protected int minActivations, maxActivations;
	protected int minViolations, maxViolations;

	private PetrinetWithAutomata net;

	private DeclarativeArc[] constraintsArray;
	private TObjectIntMap<DeclarativeArc> mapConstr2Int;

	// We cannot use MPReplayResult in the constructor as this object is constructed from within the
	// constructor of PNLogReplayProjectedVisPanel and hence the MPReplayResult object may not be available
	// just yet.
	public MPCoreInfoProvider(PetrinetWithAutomata net, Marking marking, TransEvClassMapping mapping, XLog log,
			MPReplayResult mpReplayResult) {
		super(net, marking, mapping, log, mpReplayResult.getPNRepResult());
		this.net = net;

		initializeConstraintArray(net);
		mapConstr2Int = new TObjectIntHashMap<>();
		for (int i = 0; i < constraintsArray.length; i++) {
			mapConstr2Int.put(constraintsArray[i], i);
		}

		super.initialize();
	}

	@Override
	protected void initialize() {
	}

	@Override
	public void extractInfo(boolean[] filter, Set<Integer> preservedIndexes) {
		this.minViolations = Integer.MAX_VALUE;
		this.minActivations = Integer.MAX_VALUE;
		this.maxViolations = 0;
		this.maxActivations = 0;
		this.activations = new int[constraintsArray.length];
		this.activatedTraces = new int[constraintsArray.length];
		this.violations = new int[constraintsArray.length];
		this.activationsFitting = new int[constraintsArray.length];
		// use superclass to extrace relevant info
		super.extractInfo(filter, preservedIndexes);

		// compute minima (maxima are already taken care of)
		for (int i = 0; i < activations.length; i++) {
			if (activations[i] < minActivations) {
				minActivations = activations[i];
			}
			if (violations[i] < minViolations) {
				minViolations = violations[i];
			}
		}

	}

	@Override
	protected void extractInfoFromReliableNonEmptyTrace(SyncReplayResult syncRepRes, int traceSize) {
		super.extractInfoFromReliableNonEmptyTrace(syncRepRes, traceSize);
		extractInfoOnConstraints((MPSyncReplayResult) syncRepRes, traceSize);

	}

	@Override
	protected boolean isFitting(SyncReplayResult syncRepRes, int index) {
		switch (syncRepRes.getStepTypes().get(index)) {
			case LMGOOD :
			case MINVI :
				// a fitting case only if no constraint was violated.
				return ((MPSyncReplayResult) syncRepRes).getViolatedConstraints().get(index).isEmpty();
			default :
				return false;
		}
	}

	public void extractInfoOnConstraints(MPSyncReplayResult alignment, int representedTraces) {
		//TODO: Extract info on constraint violations
		// count the activations for each globally activated constraint.
		// these constraints are activated once for each case.

		boolean fits = true;
		int[] acts = new int[activations.length];

		for (DeclarativeArc arc : net.getConstraints()) {
			if (arc.getActivationMode() == ActivationMode.GLOBAL) {
				acts[mapConstr2Int.get(arc)] += representedTraces;
			}
		}

		// go over the alignment
		for (int i = 0; i < alignment.size(); i++) {

			fits &= isFitting(alignment, i);

			if (alignment.getNodeInstance().get(i) instanceof Transition) {
				// Transition t was executed in the model. 
				Transition t = (Transition) alignment.getNodeInstance().get(i);

				// find the possibly activated constraints in the input set of t
				for (PetrinetEdge<? extends PetrinetNode, ? extends PetrinetNode> edge : net.getInEdges(t)) {
					if (edge instanceof DeclarativeArc) {
						DeclarativeArc arc = (DeclarativeArc) edge;
						int a = mapConstr2Int.get(arc);
						switch (arc.getActivationMode()) {
							case SOURCEANDTARGETACTIVATED :
							case TARGETACTIVATED :
								// t is a target of this declarativeArc. Hence
								// the arc is now activated. Count the activation
								acts[mapConstr2Int.get(arc)] += representedTraces;
								break;
							default :
								break;

						}
					}
				}

				// find the possibly activated constraints in the output set of t
				for (PetrinetEdge<? extends PetrinetNode, ? extends PetrinetNode> edge : net.getOutEdges(t)) {
					if (edge instanceof DeclarativeArc) {
						DeclarativeArc arc = (DeclarativeArc) edge;
						int a = mapConstr2Int.get(arc);
						switch (arc.getActivationMode()) {
							case SOURCEANDTARGETACTIVATED :
							case SOURCEACTIVATED :
								// t is a target of this declarativeArc. Hence
								// the arc is now activated. Count the activation
								acts[mapConstr2Int.get(arc)] += representedTraces;
								break;
							default :
								break;

						}
					}
				}

				// count the violations of constraints
				for (DeclarativeArc arc : alignment.getViolatedConstraints().get(i)) {
					int a = mapConstr2Int.get(arc);
					violations[a] += representedTraces;
					if (violations[a] > maxViolations) {
						maxViolations = violations[a];
					}

				}
			}

		}

		for (int a = 0; a < acts.length; a++) {
			activations[a] += acts[a];
			if (activations[a] > maxActivations) {
				maxActivations = activations[a];
			}
			if (acts[a] > 0) {
				activatedTraces[a] += representedTraces;
			}
			if (fits) {
				activationsFitting[a] += acts[a];
			}

		}

	}

	public int getMinimumActivations() {
		return minActivations;
	}

	public int getMaximumActivations() {
		return maxActivations;
	}

	public int getMinimumViolations() {
		return minViolations;
	}

	public int getMaximumViolations() {
		return maxViolations;
	}

	public int getViolationsFor(DeclarativeArc a) {
		return violations[mapConstr2Int.get(a)];
	}

	public int getActivatedTracesFor(DeclarativeArc a) {
		return activatedTraces[mapConstr2Int.get(a)];
	}

	public int getActivationsFor(DeclarativeArc a) {
		return activations[mapConstr2Int.get(a)];
	}

	public int getActivationsInFittingTracesFor(DeclarativeArc a) {
		return activationsFitting[mapConstr2Int.get(a)];
	}

	public int getNumConstraints() {
		return constraintsArray.length;
	}

	public DeclarativeArc[] getConstraintArray() {
		return constraintsArray;
	}

	private void initializeConstraintArray(PetrinetWithAutomata net) {
		// transition array
		List<DeclarativeArc> constraintList = new ArrayList<DeclarativeArc>(net.getConstraints());
		Collections.sort(constraintList, new Comparator<DeclarativeArc>() {

			public int compare(DeclarativeArc o1, DeclarativeArc o2) {
				return o1.toString().compareTo(o2.toString());
			}
		});
		constraintsArray = constraintList.toArray(new DeclarativeArc[constraintList.size()]);
	}

	public int[] getConstraintInfo(int arc) {
		int[] info = new int[4];
		info[0] = activations[arc];
		info[1] = activationsFitting[arc];
		info[2] = activatedTraces[arc];
		info[3] = violations[arc];
		return info;
	}

	public int getArcIndexOf(DeclarativeArc arc) {
		return mapConstr2Int.get(arc);
	}

}
